﻿Imports System.Drawing

Public Class Form1
    Private Enum CellType
        Empty
        Fish
        Shark
    End Enum

    Private Structure Cell
        Public Type As CellType
        Public BreedTimer As Integer
        Public StarveTimer As Integer ' Only for sharks
    End Structure

    Private isRunning As Boolean = False
    Private generation As Integer = 0

    Private grid(,) As Cell
    Private gridWidth As Integer = 50
    Private gridHeight As Integer = 50

    Private cellSize As Integer = 10

    Private fishBreedTime As Integer = 3

    ''TWEAK
    Private sharkBreedTime As Integer = 4
    Private sharkStarveTime As Integer = 7

    Private rand As New Random()

    Private Sub Form1_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        Me.DoubleBuffered = True
        ReDim grid(gridWidth - 1, gridHeight - 1)
        InitializeGrid()
        simulationTimer.Start()
    End Sub

    Private Sub InitializeGrid()
        For x = 0 To gridWidth - 1
            For y = 0 To gridHeight - 1
                Dim r = rand.Next(0, 10)
                If r < 3 Then
                    grid(x, y).Type = CellType.Fish
                    grid(x, y).BreedTimer = fishBreedTime
                ElseIf r = 3 Then
                    grid(x, y).Type = CellType.Shark
                    grid(x, y).BreedTimer = sharkBreedTime
                    grid(x, y).StarveTimer = sharkStarveTime
                Else
                    grid(x, y).Type = CellType.Empty
                End If
            Next
        Next
    End Sub

    Private Sub simulationTimer_Tick(sender As Object, e As EventArgs) Handles simulationTimer.Tick
        Dim newGrid(gridWidth - 1, gridHeight - 1) As Cell
        For x = 0 To gridWidth - 1
            For y = 0 To gridHeight - 1
                newGrid(x, y) = grid(x, y)
            Next
        Next

        For x = 0 To gridWidth - 1
            For y = 0 To gridHeight - 1
                Dim cell = grid(x, y)

                Select Case cell.Type
                    Case CellType.Fish
                        MoveFish(x, y, newGrid)
                    Case CellType.Shark
                        MoveShark(x, y, newGrid)
                End Select
            Next
        Next

        grid = newGrid
        generation += 1
        UpdateStats()
        canvasPanel.Invalidate()
    End Sub

    Private Sub MoveFish(x As Integer, y As Integer, ByRef newGrid(,) As Cell)
        Dim neighbors = GetEmptyNeighbors(x, y)
        If neighbors.Count > 0 Then
            Dim index = rand.Next(neighbors.Count)
            Dim nx = neighbors(index).Item1
            Dim ny = neighbors(index).Item2

            newGrid(nx, ny) = grid(x, y) ' Move the fish to the new location
            newGrid(nx, ny).BreedTimer = grid(x, y).BreedTimer - 1

            If grid(x, y).BreedTimer <= 0 Then
                newGrid(x, y).Type = CellType.Fish
                newGrid(x, y).BreedTimer = fishBreedTime
                newGrid(nx, ny).BreedTimer = fishBreedTime
            Else
                newGrid(x, y).Type = CellType.Empty
            End If
        Else
            newGrid(x, y).BreedTimer -= 1
        End If
    End Sub

    Private Sub MoveShark(x As Integer, y As Integer, ByRef newGrid(,) As Cell)
        Dim fishNeighbors = GetNeighborsOfType(x, y, CellType.Fish)
        Dim shark = grid(x, y)
        Dim moved = False
        Dim targetX = x
        Dim targetY = y

        If fishNeighbors.Count > 0 Then
            Dim index = rand.Next(fishNeighbors.Count)
            targetX = fishNeighbors(index).Item1
            targetY = fishNeighbors(index).Item2

            shark.StarveTimer = sharkStarveTime ' Ate fish
            shark.BreedTimer -= 1
            newGrid(targetX, targetY) = shark

            If shark.BreedTimer <= 0 Then
                newGrid(x, y).Type = CellType.Shark
                newGrid(x, y).BreedTimer = sharkBreedTime
                newGrid(x, y).StarveTimer = sharkStarveTime
            Else
                newGrid(x, y).Type = CellType.Empty
            End If
            moved = True

        Else
            Dim emptyNeighbors = GetEmptyNeighbors(x, y)
            If emptyNeighbors.Count > 0 Then
                Dim index = rand.Next(emptyNeighbors.Count)
                targetX = emptyNeighbors(index).Item1
                targetY = emptyNeighbors(index).Item2

                shark.StarveTimer -= 1
                shark.BreedTimer -= 1
                newGrid(targetX, targetY) = shark

                If shark.BreedTimer <= 0 Then
                    newGrid(x, y).Type = CellType.Shark
                    newGrid(x, y).BreedTimer = sharkBreedTime
                    newGrid(x, y).StarveTimer = sharkStarveTime
                Else
                    newGrid(x, y).Type = CellType.Empty
                End If
                moved = True
            End If
        End If

        ' Final starvation check
        If Not moved Then
            newGrid(x, y).StarveTimer -= 1
            If newGrid(x, y).StarveTimer <= 0 Then
                newGrid(x, y).Type = CellType.Empty
            End If
        Else
            If newGrid(targetX, targetY).StarveTimer <= 0 Then
                newGrid(targetX, targetY).Type = CellType.Empty
            End If
        End If
    End Sub

    Private Function GetEmptyNeighbors(x As Integer, y As Integer) As List(Of (Integer, Integer))
        Return GetNeighborsOfType(x, y, CellType.Empty)
    End Function

    Private Function GetNeighborsOfType(x As Integer, y As Integer, cellType As CellType) As List(Of (Integer, Integer))
        Dim neighbors As New List(Of (Integer, Integer))()
        Dim dx() = {-1, 1, 0, 0}
        Dim dy() = {0, 0, -1, 1}
        For i = 0 To 3
            Dim nx = (x + dx(i) + gridWidth) Mod gridWidth
            Dim ny = (y + dy(i) + gridHeight) Mod gridHeight
            If grid(nx, ny).Type = cellType Then
                neighbors.Add((nx, ny))
            End If
        Next
        Return neighbors
    End Function

    Private Sub canvasPanel_Paint(sender As Object, e As PaintEventArgs) Handles canvasPanel.Paint
        Dim g = e.Graphics
        For x = 0 To gridWidth - 1
            For y = 0 To gridHeight - 1
                Dim color As Color
                Select Case grid(x, y).Type
                    Case CellType.Fish
                        color = Color.LightBlue
                    Case CellType.Shark
                        color = Color.Red
                    Case Else
                        color = Color.Black
                End Select
                Using brush As New SolidBrush(color)
                    g.FillRectangle(brush, x * cellSize, y * cellSize, cellSize, cellSize)
                End Using
            Next
        Next
    End Sub

    Private Sub btnStartPause_Click(sender As Object, e As EventArgs) Handles btnStartPause.Click
        isRunning = Not isRunning
        If isRunning Then
            simulationTimer.Start()
            btnStartPause.Text = "Pause"
        Else
            simulationTimer.Stop()
            btnStartPause.Text = "Start"
        End If
    End Sub

    Private Sub btnReset_Click(sender As Object, e As EventArgs) Handles btnReset.Click
        simulationTimer.Stop()
        btnStartPause.Text = "Start"
        isRunning = False

        gridWidth = CInt(numGridSize.Value)
        gridHeight = gridWidth
        cellSize = Math.Max(4, 600 \ gridWidth)
        ReDim grid(gridWidth - 1, gridHeight - 1)
        InitializeGrid()
        canvasPanel.Invalidate()
        generation = 0
        UpdateStats()
    End Sub

    Private Sub trackSpeed_Scroll(sender As Object, e As EventArgs) Handles trackSpeed.Scroll
        simulationTimer.Interval = trackSpeed.Value
    End Sub

    Private Sub UpdateStats()
        Dim fishCount As Integer = 0
        Dim sharkCount As Integer = 0

        For x = 0 To gridWidth - 1
            For y = 0 To gridHeight - 1
                Select Case grid(x, y).Type
                    Case CellType.Fish : fishCount += 1
                    Case CellType.Shark : sharkCount += 1
                End Select
            Next
        Next

        lblStats.Text = $"Gen: {generation} | Fish: {fishCount} | Sharks: {sharkCount}"
    End Sub

    Private Sub Form1_Resize(sender As Object, e As EventArgs) Handles Me.Resize
        canvasPanel.Width = gridWidth * cellSize + 2
        canvasPanel.Height = gridHeight * cellSize + 2
    End Sub

End Class
